from kafkacrypto.exceptions import KafkaCryptoMessageError

class KafkaCryptoMessage(object):
  """Class encapsulating crypto-handled KafkaCrypto messages. 

  Keyword Arguments:
              msg (bytes): Message to turn into a KafkaCryptoMessage object
      ipt (bool,optional): Is message plaintext, or ciphertext? 
                           Default: plaintext.
  """
  #
  # Per instance, defined in init
  # _msg: message bytes
  # _ipt: plainext boolean
  #
  def __init__(self, msg, ipt=True):
    if (not isinstance(msg, (bytes,bytearray))):
      raise KafkaCryptoMessageError("Passed message not bytes-like!")
    if (not isinstance(ipt, (bool))):
      raise KafkaCryptoMessageError("Passed plaintext flag not a boolean!")
    self._msg = msg
    self._ipt = ipt

  def isCleartext(self):
    return self._ipt

  def getMessage(self):
    if (not self.isCleartext()):
      raise KafkaCryptoMessageError("Message not decrypted!")
    return self._msg

  def __bytes__(self):
    return self._msg

  @staticmethod
  def fromBytes(rawmsg):
    if (not isinstance(rawmsg, (bytes,bytearray))):
      raise KafkaCryptoMessageError("Passed raw message not bytes-like!")
    if (rawmsg[0] == 0):
      return KafkaCryptoMessage(rawmsg[1:],True)
    else:
      return KafkaCryptoMessage(rawmsg,False)



from kafka import KafkaConsumer as Consumer, KafkaProducer, TopicPartition
from collections import namedtuple

TopicPartitionOffset = namedtuple("TopicPartitionOffset",
  ["topic", "partition", "offset"])

class KafkaConsumer(Consumer):
  def assign_and_seek(self, partoffs):
    tps = []
    for tpo in partoffs:
      tps.append(TopicPartition(tpo.topic,tpo.partition))
    super().assign(tps)
    for tpo in partoffs:
      if (tpo.offset > 0):
        super().seek(TopicPartition(tpo.topic,tpo.partition),tpo.offset)



import pysodium
import msgpack
from binascii import unhexlify, hexlify
from time import time
from kafkacrypto.exceptions import KafkaCryptoUtilError

class PasswordProvisioner(object):
  """The PasswordProvisioner class instantiates three provisioner keys, and uses them
  to sign signing keys of new producers/consumers.

  Alternative versions using, e.g. security keys or similar, can also be written.

  Keyword Arguments:
  password (str): Password from which to derive the provisioning secret keys.

  """
  # Constrained devices cannot use larger numbers than interactive
  _ops = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
  _mem = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE

  def __init__(self, password, rot):
    if (isinstance(password,(str,))):
      password = password.encode('utf-8')
    try:
      rot = unhexlify(rot)
    except:
      pass
    self._salt = {}
    self._salt['producer'] = pysodium.crypto_hash_sha256(b'producer' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._salt['consumer'] = pysodium.crypto_hash_sha256(b'consumer' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._salt['prodcon'] = pysodium.crypto_hash_sha256(b'prodcon' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._seed = {}
    self._pk = {}
    self._sk = {}
    print("")
    print("Deriving provisioning keys with:")
    print("  opsl = ", self._ops)
    print("  meml = ", self._mem)
    for key in self._salt.keys():
      print("  salt (", key, ") = ", hexlify(self._salt[key]))
      self._seed[key] = pysodium.crypto_pwhash_scryptsalsa208sha256(pysodium.crypto_sign_SEEDBYTES, password, self._salt[key], opslimit=self._ops, memlimit=self._mem)
      self._pk[key], self._sk[key] = pysodium.crypto_sign_seed_keypair(self._seed[key])
      print("  Signing Public Key (", key, "): ", hexlify(self._pk[key]))
    print("")

class PasswordROT(object):
  """The PasswordROT class instantiates a root of trust keypair.

  Alternative versions using, e.g. security keys or similar, can also be written.

  Keyword Arguments:
  password (str): Password from which to derive the secret key.
  """
  # Constrained devices cannot use larger numbers than interactive
  _ops = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
  _mem = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE

  def __init__(self, password):
    if (isinstance(password,(str,))):
      password = password.encode('utf-8')
    self._salt = pysodium.crypto_hash_sha256(b'Root of Trust' + password)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    print("")
    print("Deriving root key with:")
    print("  opsl = ", self._ops)
    print("  meml = ", self._mem)
    print("  salt = ", hexlify(self._salt))
    self._seed = pysodium.crypto_pwhash_scryptsalsa208sha256(pysodium.crypto_sign_SEEDBYTES, password, self._salt, opslimit=self._ops, memlimit=self._mem)
    self._pk, self._sk = pysodium.crypto_sign_seed_keypair(self._seed)
    print("  Root Public Key: ", hexlify(self._pk))
    print("")



from threading import Lock
import inspect
import msgpack
import logging
from kafkacrypto import KafkaProducer, KafkaConsumer
from kafkacrypto.exceptions import KafkaCryptoBaseError
from kafkacrypto.cryptokey import CryptoKey

class KafkaCryptoBase(object):
  """Base class for the handling the sending and receiving of encrypted messages.

  Keyword Arguments:
              nodeID (str): Node ID
        kp (KafkaProducer): Pre-initialized KafkaProducer, ready for
                            handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
        kc (KafkaConsumer): Pre-initialized KafkaConsumer, ready for      
       	       	       	    handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
         config (str,dict): Either a filename in which configuration
                            data is a stored, or a dict of config
                            parameters. Set to None to load from the
                            default location based on nodeID.
       cryptokey (str,obj): Either a filename in which the crypto
                            private key is stored, or an object 
                            implementing the necessary functions
                            (encrypt_keys, decrypt_keys, sign_spk,
                            load/store_crypto_opaque).
                            Set to None to use the default file
                            location based on nodeID.
  """

  #
  # Global configuration defaults
  #
  DEFAULTS = { 'TOPIC_SEPARATOR': b'.',        # separator of topic name components, used to find root name and subs/keys
               'TOPIC_SUFFIX_SUBS': b'.subs',  # suffixes should begin with separator or things will not work!
               'TOPIC_SUFFIX_KEYS': b'.keys',
               'TOPIC_SUFFIX_REQS': b'.reqs',
               'CRYPTO_MAX_PGEN_AGE': 604800,  # in s
               'CRYPTO_SUB_INTERVAL': 60,      # in s
               'CRYPTO_RATCHET_INTERVAL': 86400,  # in s
               'MGMT_POLL_INTERVAL': 500,      # in ms
               'MGMT_POLL_RECORDS': 8,         # poll fetches by topic-partition. So limit number per call to sample all tps
               'MGMT_SUBSCRIBE_INTERVAL': 300, # in sec
               'MGMT_LONG_KEYINDEX': False,
             }

  def __init__(self, nodeID, kp, kc, config, cryptokey):
    if (not isinstance(nodeID, (str)) or len(nodeID) < 1):
      raise KafkaCryptoBaseError("Node ID " + str(nodeID) + " not a string or not specified!")
    if (not isinstance(kp, (KafkaProducer))):
      raise KafkaCryptoBaseError("Invalid Kafka Producer supplied!")
    if (not isinstance(kc, (KafkaConsumer))):
      raise KafkaCryptoBaseError("Invalid Kafka Consumer supplied!")
    if (config is None):
      config = nodeID + ".config"
    if (isinstance(config,(str,))):
      try:
        with open(config, 'rb') as config_file:
          config = msgpack.unpackb(config_file.read())
      except:
        config = {}
    self.__config(config)
    if (cryptokey is None):
      cryptokey = nodeID + ".crypto"
    if (isinstance(cryptokey,(str))):
      cryptokey = CryptoKey(file=cryptokey)
    if (not hasattr(cryptokey, 'decrypt_keys') or not inspect.isroutine(cryptokey.decrypt_keys) or not hasattr(cryptokey, 'encrypt_keys') or not inspect.isroutine(cryptokey.encrypt_keys) or
        not hasattr(cryptokey, 'signed_epk') or not inspect.isroutine(cryptokey.signed_epk) or not hasattr(cryptokey, 'load_crypto_opaque') or not inspect.isroutine(cryptokey.load_crypto_opaque) or
        not hasattr(cryptokey, 'store_crypto_opaque') or not inspect.isroutine(cryptokey.store_crypto_opaque)):
      raise KafkaCryptoBaseError("Invalid cryptokey source supplied!")

    self._nodeID = nodeID
    self._lock = Lock()
    self._kp = kp
    self._kc = kc
    self._cryptokey = cryptokey 
    self._logger = logging.getLogger(__name__)

  def __config(self, config):
    # config should be a map with names that we can override defaults with.
    for k in self.DEFAULTS.keys():
      v = self.DEFAULTS[k]
      if k in config.keys():
        v = config[k]
      elif (isinstance(k,(str,)) and k.encode('utf-8') in config.keys()):
        v = config[k.encode('utf-8')]
      setattr(self, k, v)

  def get_root(self,topic):
    root = topic.index(self.TOPIC_SEPARATOR)
    if (root != -1):
      root = topic[0:root]
    else:
      root = topic
    return root



from time import time
import traceback
import pysodium
import msgpack
from kafkacrypto.chain import process_chain

class Provisioners(object):
  """Class utilizing file-backed storage to validate key requests

  Keyword Arguments:
          file (str,file): Filename or File IO object for storing provisioners.
                           Must be seekable, with read/write permission, and
                           honor sync requests.
  """
  #
  # Per instance, defined in init
  #      __file: File object
  #      __spks: list of allowed provisioner public keys
  #
  def __init__(self, file):
    if (isinstance(file, (str))):
      file = open(file, 'rb', 0)
    self.__file = file
    self.__file.seek(0,0)
    contents = msgpack.unpackb(self.__file.read())
    self.__spks = contents
    self.__file.close()

  def reencrypt_request(self, topic, cryptokey, msgkey=None, msgval=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    #
    # msgval should be a msgpacked chain chaining to a provisioner.
    # The last item in the array is the encrypted public key to retransmit.
    #
    try:
      pk = None
      for prov in self.__spks:
        try:
          pk = process_chain(topic,prov,msgval,b'key-encrypt-subscribe')
        except:
          pass
      if (prov is None):
        raise ValueError("Chain not signed by authorized provisioner.")
      tmp, msg = cryptokey.signed_epk(topic, epk=pk[2])
    except Exception as e:
      print("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      return (None, None)
    return (msgkey, msg)



from time import time
import traceback
import pysodium
import msgpack
import logging
from kafkacrypto.chain import process_chain

class CryptoKey(object):
  """Class utilizing file-backed storage to encrypt and decrypt encryption keys
     to/from group members.

  Keyword Arguments:
          file (str,file): Filename or File IO object for storing crypto info.
                           Must be seekable, with read/write permission, and
                           honor sync requests.
  """
  RANDOM_BYTES = 32
  #
  # Per instance, defined in init
  #      __file: File object
  #       __rot: Root of Trust for received messages
  #  __chainrot: Root of Trust for our trust chain
  #    __maxage: Maximum age (seconds)
  # __spk_chain: signing public key chain to root of trust, as array
  #       __spk: signing public key
  #       __ssk: signing private (secret) key
  # Generated ephemerially, on demand:
  #       __epk: dict of encrypting public keys (by topic)
  #       __esk: dict of encrypting private (secret) keys (by topic)
  #
  def __init__(self, file):
    self._logger = logging.getLogger(__name__)
    if (isinstance(file, (str))):
      file = open(file, 'rb+', 0)
    self.__file = file
    self.__file.seek(0,0)
    data = self.__file.read()
    datalen = len(data)
    contents = None
    while len(data) and contents is None:
      try:
        contents = msgpack.unpackb(data)
      except msgpack.exceptions.ExtraData:
        data = data[:-1]
    if len(data) != datalen:
      self._logger.warning("Cryptokey file had extraneous bytes at end, attempting load anyways.")
    self.__maxage = contents[0]
    self.__rot = contents[1]
    self.__chainrot = contents[2]
    self.__esk = {}
    self.__epk = {}
    self.__ssk = contents[3]
    self.__spk = pysodium.crypto_sign_sk_to_pk(self.__ssk)
    self.__spk_chain = []
    self.__crypto_opaque = msgpack.packb([])
    if (len(contents) > 5):
      self.__crypto_opaque = contents[5]
    # update chain last (since it might rewrite file)
    self.__update_spk_chain(contents[4])

  def encrypt_keys(self, keyidxs, keys, topic, msgkey=None, msgval=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    #
    # msgval should be a msgpacked chain.
    # The public key in the array is the public key to send the key to, using a
    # common DH-derived key between that public key and our private encryption key.
    # Then there is at least one more additional item, random bytes:
    # (3) random bytes
    # There then might be additional public keys:
    # (4) public key
    # ...
    # Which, if present, are multiplied by our secret key and returned along
    # with our public key in the response. This is only safe because:
    #  1. A fresh ephemeral secret key is used for each call of encrypt_keys
    #  2. We validate that the additional public keys are not equal to the
    #     one used to derive the shared secret for this call
    # Together, these ensure that an attacker cannot use the additional public
    # keys feature to learn the common DH shared secret for another session.
    #
    try:
      pk = process_chain(topic,self.__rot,msgval,b'key-encrypt-request')
      # Construct shared secret as sha256(topic || random0 || random1 || our_private*their_public)
      self.__generate_esk(topic)
      random1 = pysodium.randombytes(self.RANDOM_BYTES)
      ss = pysodium.crypto_hash_sha256(topic + pk[3] + random1
           + pysodium.crypto_scalarmult_curve25519(self.__esk[topic],pk[2]))[0:pysodium.crypto_secretbox_KEYBYTES]
      nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
      # encrypt keys and key indexes (MAC appended, nonce prepended)
      msg = []
      for i in range(0,len(keyidxs)):
        msg.append(keyidxs[i])
        msg.append(keys[i])
      msg = msgpack.packb(msg)
      msg = nonce + pysodium.crypto_secretbox(msg,nonce,ss)
      # this is then put in a msgpack array with the appropriate max_age, poison, and public key(s)
      poison = msgpack.packb([[b'topics',[topic]],[b'usages',[b'key-encrypt']]])
      pks = [self.__epk[topic]]
      for extrapk in pk[4:]:
        if (extrapk != pk[3]):
          pks.append(pysodium.crypto_scalarmult_curve25519(self.__esk[topic],extrapk))
      msg = msgpack.packb([time()+self.__maxage,poison,pks,[pk[3],random1],msg])
      # and signed with our signing key
      msg = pysodium.crypto_sign(msg, self.__ssk)
      # and finally put as last member of a msgpacked array chaining to ROT
      tchain = self.__spk_chain.copy()
      tchain.append(msg)
      msg = msgpack.packb(tchain)
      # remove ephemeral key
      self.__remove_esk(topic)
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      return (None, None)
    return (msgkey, msg)

  def decrypt_keys(self, topic, msgkey=None, msgval=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    #
    # msgval should be a msgpacked chain.
    # The public key in the array is a set of public key(s) to combine with our
    # encryption key to get the secret to decrypt the key. If we do not
    # have an encryption key for the topic, we cannot do it until we have one.
    # The next item is then the pair of random values for generating the shared
    # secret, followed by the actual key message.
    #
    try:
      if (not topic in self.__esk.keys()):
        raise ValueError
      pk = process_chain(topic,self.__rot,msgval,b'key-encrypt')
      if (len(pk) < 5):
        raise ValueError
      random0 = pk[3][0]
      random1 = pk[3][1]
      nonce = pk[4][0:pysodium.crypto_secretbox_NONCEBYTES]
      msg = pk[4][pysodium.crypto_secretbox_NONCEBYTES:]
      for cpk in pk[2]:
        # Construct candidate shared secrets as sha256(topic || our_private*their_public)
        ss = pysodium.crypto_hash_sha256(topic + random0 + random1
             + pysodium.crypto_scalarmult_curve25519(self.__esk[topic],cpk))[0:pysodium.crypto_secretbox_KEYBYTES]
        # decrypt and return key
        try:
          msg = msgpack.unpackb(pysodium.crypto_secretbox_open(msg,nonce,ss))
          rvs = {}
          for i in range(0,len(msg),2):
            rvs[msg[i]] = msg[i+1]
          if len(rvs) < 1 or 2*len(rvs) != len(msg):
            raise ValueError
        except:
          pass
        else:
          self.__remove_esk(topic)
          return rvs
      raise ValueError
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      pass
    return None

  def signed_epk(self, topic, epk=None):
    #
    # returns the public key of the current encryption key for the specified topic
    # (generating a new one if not present), signed by our signing key,
    # with a fresh random value, and with the chain to the ROT prepended.
    #
    try:
      if not (epk is None) and isinstance(epk,(bytes,bytearray)):
        self.__epk[topic] = epk
      if not (topic in self.__epk):
        self.__generate_esk(topic)
      random0 = pysodium.randombytes(self.RANDOM_BYTES)
      # we allow either direct-to-producer or via-controller key establishment
      poison = msgpack.packb([[b'topics',[topic]],[b'usages',[b'key-encrypt-request',b'key-encrypt-subscribe']]])
      msg = msgpack.packb([time()+self.__maxage,poison,self.__epk[topic],random0])
      # and signed with our signing key
      msg = pysodium.crypto_sign(msg, self.__ssk)
      # and finally put as last member of a msgpacked array chaining to ROT
      tchain = self.__spk_chain.copy()
      tchain.append(msg)
      msg = msgpack.packb(tchain)
      return (self.__spk, msg)
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      pass
    return (None, None)

  def store_crypto_opaque(self, crypto_opaque):
    self.__crypto_opaque = msgpack.packb([crypto_opaque])
    self.__update_file()

  def load_crypto_opaque(self):
    v = msgpack.unpackb(self.__crypto_opaque)
    if (len(v) > 0):
      return v[0]
    else:
      return None

  def __generate_esk(self, topic):
    # ephemeral keys are use once only, so always ok to overwrite
    self.__esk[topic] = pysodium.randombytes(pysodium.crypto_scalarmult_curve25519_BYTES)
    self.__epk[topic] = pysodium.crypto_scalarmult_curve25519_base(self.__esk[topic])

  def __remove_esk(self, topic):
    self.__esk.pop(topic)
    self.__epk.pop(topic)

  def __update_file(self):
    self.__file.seek(0,0)
    self.__file.write(msgpack.packb([self.__maxage, self.__rot, self.__chainrot, self.__ssk, msgpack.packb(self.__spk_chain), self.__crypto_opaque]))
    self.__file.flush()
    self.__file.truncate()
    self.__file.flush()

  def __update_spk_chain(self, newchain):
    #
    # We have a new candidate chain to replace the current one (if any). We
    # first must check it is a valid chain ending in our signing public key.
    #
    try:
      pk = process_chain(b'',self.__chainrot,newchain,b'')
      if (len(pk) < 3 or pk[2] != self.__spk):
        raise ValueError
      # If we get here, it is a valid chain. So now we need
      # to see if it is "superior" than our current chain.
      # This means a larger minimum max_age.
      min_max_age = 0
      for cpk in self.__spk_chain:
        if (min_max_age == 0 or cpk[0]<min_max_age):
          min_max_age = cpk[0]
      for cpk in newchain:
        if (min_max_age != 0 and cpk[0]<min_max_age):
          raise ValueError
      self.__spk_chain = msgpack.unpackb(newchain)
      self.__update_file()
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      pass



from kafkacrypto.exceptions import KafkaCryptoGeneratorError
import pysodium

class KeyGenerator(object):
  """Class implementing a blake2b-based key/nonce generator

  Keyword Arguments:
          secret (bytes, optional): 32-byte Secret used to generate keys
             ctx (bytes, optional): 16-byte context used to generate keys
  """
  #
  # Generator global configuration.
  #
  SECRETSIZE = 32
  KEYSIZE = 32
  NONCESIZE = 24
  SALTSIZE = 16
  MSG = b''

  #
  # Per instance, defined in init
  # __secret: Secret used to generate keys
  #    __ctx: Default context used to generate keys
  #   __salt: Default salt, incremented after each invocation
  #
  def __init__(self, secret=b'\x00'*SECRETSIZE, ctx=b'generator'+(b'\x00'*7)):
    if (not isinstance(ctx, (bytes,bytearray))):
      raise KafkaCryptoGeneratorError("Context is not bytes!")
    self.__secret = None
    self.__ctx = ctx
    self.rekey(secret)

  def rekey(self, secret):
    if (not isinstance(secret, (bytes,bytearray)) or len(secret) != self.SECRETSIZE or (not (self.__secret is None) and secret == self.__secret)):
      raise KafkaCryptoGeneratorError("Secret is malformed!")
    self.__secret = secret
    self.__salt = b'\x00' * self.SALTSIZE

  def generate(self,msg=MSG,ctx=None,salt=None,keysize=KEYSIZE,noncesize=NONCESIZE):
    # pysodium silently computes the hash of an empty string if input is not bytes, so check for
    # and catch that.
    if (ctx is None or not isinstance(ctx,(bytes,bytearray))):
      ctx = self.__ctx
    if (salt is None or not isinstance(salt,(bytes,bytearray))):
      salt = self.__salt
    if (not isinstance(msg,(bytes,bytearray))):
      raise KafkaCryptoGeneratorError("Message is not bytes!")
    keynonce = pysodium.crypto_generichash_blake2b_salt_personal(msg,key=self.__secret,salt=salt,personal=ctx,outlen=keysize+noncesize)
    self.__salt = (int.from_bytes(self.__salt, byteorder='little', signed=False)+1).to_bytes(self.SALTSIZE, byteorder='little', signed=False)
    return (keynonce[0:keysize], keynonce[-noncesize:])

  def salt(self):
    return self.__salt

  @staticmethod
  def get_key_value_generators(secret):
    return (KeyGenerator(secret=secret,ctx=b'key'+(b'\x00'*13)), KeyGenerator(secret=secret,ctx=b'value'+(b'\x00'*11)))



from time import time
import pysodium
import msgpack

def process_chain(topic, rot, chain, usage):
  #
  # chain should be a single msgpack array, with the next item in the array
  # signed by the entity with public key given in the current item. The first
  # item must be signed by the root of trust. Each (signed) item is itself a
  # msgpack array containing:
  # (0) a max_age timestamp (64-bit unsigned unix maximum time signature is valid)
  # (1) a poison array. Sets allowed signature key uses
  # (2) public key
  # Possibly more items in (3)+, especially common for the final item,
  # but not needed to check the chain
  #
  if (len(rot) > 0):
    val = [rot] + msgpack.unpackb(chain)
  else:
    val = msgpack.unpackb(chain)
  pk = None
  for npk in val:
    if not (pk is None):
      pk = msgpack.unpackb(pysodium.crypto_sign_open(npk, pk[2]))
    else:
      pk = msgpack.unpackb(npk) # root is unsigned
    if (len(pk) < 3 or (pk[0]<time() and pk[0]>0)):
      raise ValueError("Invalid Signature on Chain")
    poison = msgpack.unpackb(pk[1])
    for ku in poison:
      if (ku[0] == b'topics'):
        if (not (topic in ku[1]) and len(topic)>0):
          raise ValueError("Chain signature used for disallowed topic")
      elif (ku[0] == b'usages'):
        if (not (usage in ku[1]) and len(usage)>0):
          raise ValueError("Chain signature used for disallowed usage")
      else:
        raise ValueError("Unknown chain signature poison")
  return pk



name='kafkacrypto'
# version and license information in setup.py
__all__ = []

try:
  import confluent_kafka
  from kafkacrypto.confluent_kafka_wrapper import KafkaConsumer,KafkaProducer,TopicPartition,TopicPartitionOffset
except ImportError:
  # fallback to kafka-python
  from kafkacrypto.kafka_python_wrapper import KafkaConsumer,KafkaProducer,TopicPartition,TopicPartitionOffset

__all__.extend(['KafkaConsumer', 'KafkaProducer', 'TopicPartition', 'TopicPartitionOffset'])

from kafkacrypto.message import KafkaCryptoMessage
from kafkacrypto.crypto import KafkaCrypto
from kafkacrypto.controller import KafkaCryptoController
__all__.extend([ 'KafkaCryptoMessage', 'KafkaCrypto', 'KafkaCryptoController'])




import copy
import inspect
import logging
from confluent_kafka import Producer, Consumer, TopicPartition as TopicPartitionOffset, OFFSET_BEGINNING
from kafkacrypto.exceptions import KafkaCryptoWrapperError
from collections import namedtuple

TopicPartition = namedtuple("TopicPartition", ["topic", "partition"])
Message = namedtuple("Message",
  ["topic", "partition", "offset", "timestamp", "headers", "key", "value"])

class KafkaConsumer(Consumer):
  """
  Thin wrapper around confluent-kafka providing methods unique to kafka-python
  to allow them to be used interchangeably. Unless otherwise noted, parameters should
  be supplied as expected by the underlying confluent_kafka classes. Use at your own risk!
  """
  CONFIG_MAP = { 'ssl_cafile': 'ssl.ca.location',
                 'ssl_certfile': 'ssl.certificate.location',
                 'ssl_keyfile': 'ssl.key.location',
                 'ssl_crlfile': 'ssl.crl.location',
                 'ssl_password': 'ssl.key.password',
                 'ssl_ciphers': 'ssl.cipher.suites',
                 'max_partition_fetch_bytes': 'message.max.bytes', # this maps the hard limit functionality appropriately
                 'bootstrap_servers': 'bootstrap.servers',
                 'client_id': 'client.id',
                 'group_id': 'group.id',
                 'fetch_min_bytes': 'fetch.min.bytes',
                 'fetch_max_wait_ms': 'fetch.wait.max.ms',
                 'fetch_max_bytes': 'fetch.max.bytes',
                 'request_timeout_ms': 'request.timeout.ms',
                 'retry_backoff_ms': 'retry.backoff.ms',
                 'reconnect_backoff_ms': 'reconnect.backoff.ms',
                 'reconnect_backoff_max_ms': 'reconnect.backoff.max.ms',
                 'max_in_flight_requests_per_connection': 'max.in.flight.requests.per.connection',
                 'auto_offset_reset': 'auto.offset.reset',
                 'enable_auto_commit': 'enable.auto.commit',
                 'auto_commit_interval_ms': 'auto.commit.interval.ms',
                 'check_crcs': 'check.crcs',
                 'metadata_max_age_ms': 'metadata.max.age.ms',
                 'partition_assignment_strategy': 'partition.assignment.strategy',
                 'max_poll_interval_ms': 'max.poll.interval.ms',
                 'session_timeout_ms': 'session.timeout.ms',
                 'heartbeat_interval_ms': 'heartbeat.interval.ms',
                 'receive_buffer_bytes': 'socket.receive.buffer.bytes',
                 'send_buffer_bytes': 'socket.send.buffer.bytes',
                 'security_protocol': 'security.protocol',
                 'sasl_mechanism': 'sasl.mechanism',
                 'sasl_plain_username': 'sasl.username',
                 'sasl_plain_password': 'sasl.password',
                 'sasl_kerberos_service_name': 'sasl.kerberos.service.name',
               }
  CONFIG_MAP_LOCAL = { 'max_poll_records': 500,
                     }
  CONFIG_MAP_NULL = [ 'ssl_context', 
                      'socket_options', 
                      'default_offset_commit_callback', 
                      'consumer_timeout_ms',
                      'ssl_check_hostname',
                      'api_version',
                      'api_version_auto_timeout_ms',
                      'connections_max_idle_ms',
                      'metrics_reporters',
                      'metrics_num_samples',
                      'metrics_sample_window_ms',
                      'selector',
                      'exclude_internal_topics',
                      'sasl_kerberos_domain_name',
                      'sasl_oauth_token_provider',
                    ]
  def __init__(self, *topics, **configs):
    self._log = logging.getLogger(__name__)
    self.raw_config = configs
    self.cf_config = {}
    self.config = {}
    self.kds = lambda topic, _bytes: _bytes
    self.vds = lambda topic, _bytes: _bytes
    for k in self.CONFIG_MAP_LOCAL.keys():
      self.config[k] = self.CONFIG_MAP_LOCAL[k]
    for k in self.raw_config.keys():
      if (k == 'key_deserializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'deserialize') or not inspect.isroutine(self.raw_config[k].deserialize)):
          self.kds = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.kds = self.raw_config[k].deserialize
        self.config[k] = self.kds
      elif (k == 'value_deserializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'deserialize') or not inspect.isroutine(self.raw_config[k].deserialize)):
          self.vds = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.vds = self.raw_config[k].deserialize
        self.config[k] = self.vds
      elif (k in self.CONFIG_MAP.keys()):
        if not (self.raw_config[k] is None):
          self.cf_config[self.CONFIG_MAP[k]] = self.raw_config[k]
      elif (k in self.CONFIG_MAP_LOCAL):
        self.config[k] = self.raw_config[k]
      elif (k in self.CONFIG_MAP_NULL):
        self._log.warning("Warning: Unsupported kafka-python parameter passed to confluent wrapper: %s, %s", k, self.raw_config[k])
      else:
        self.cf_config[k] = self.raw_config[k]
    for oldk,newk in self.CONFIG_MAP.items():
      if newk in self.cf_config.keys():
        self.config[oldk] = self.cf_config[newk]
    super().__init__(self.cf_config)
    if topics:
      self.subscribe(topics)

  def subscribe(self, topics=None, pattern=None, listener=None):
    if not topics is None:
      topics = copy.copy(topics)
    else:
      topics = []
    if not pattern is None:
      if pattern[0] != '^':
        topics.append('^' + pattern)
      else:
        topics.append(pattern)
    if listener is not None:
      return super().subscribe(topics, listener=listener)
    else:
      return super().subscribe(topics)

  def poll(self, timeout_ms=0, max_records=None):
    rvm = {}
    msgs = super().consume(self.config['max_poll_records'],timeout_ms/1000.0)
    for msg in msgs:
      if not (msg is None) and msg.error() is None:
        rvk = TopicPartition(msg.topic(),msg.partition())
        rv = Message(rvk.topic, rvk.partition, msg.offset(), msg.timestamp(), msg.headers(), self.kds(rvk.topic,msg.key()), self.vds(rvk.topic,msg.value()))
        if rvk in rvm.keys():
          rvm[rvk].append(rv)
        else:
          rvm[rvk] = [rv]
    return rvm

  def assign(self, partitions):
    tps = []
    for tp in partitions:
      tps.append(TopicPartitionOffset(tp.topic, tp.partition))
    return super().assign(tps)

  def assign_and_seek(self, partoffs):
    tps = []
    for tpo in partoffs:
      if (tpo.offset > 0):
        tps.append(TopicPartitionOffset(tpo.topic, tpo.partition, tpo.offset))
      else:
        tps.append(TopicPartitionOffset(tpo.topic, tpo.partition))
    return super().assign(tps)

  def seek(self, tp, offset):
    return super().seek(TopicPartitionOffset(tp.topic, tp.partition, offset))

  def seek_to_beginning(self):
    for tp in self.assignment():
      try:
        self.seek(tp, OFFSET_BEGINNING)
      except:
        pass

  def __iter__(self):
    return self

  def __next__(self):
    rv = self.poll(timeout_ms=-1000.0,max_records=1)
    if len(rv) == 1:
      return list(rv.values())[0][0]
    raise StopIteration("Poll Failed!")

class KafkaProducer(Producer):
  """
  Thin wrapper around confluent-kafka providing methods unique to kafka-python
  to allow them to be used interchangeably. Unless otherwise noted, parameters should
  be supplied as expected by the underlying confluent_kafka classes. Use at your own risk!
  """
  CONFIG_MAP = { 'ssl_cafile': 'ssl.ca.location',
                 'ssl_certfile': 'ssl.certificate.location',
                 'ssl_keyfile': 'ssl.key.location',
                 'ssl_crlfile': 'ssl.crl.location',
                 'ssl_password': 'ssl.key.password',
                 'ssl_ciphers': 'ssl.cipher.suites',
                 'max_request_size': 'message.max.bytes', # this maps the hard limit functionality appropriately                                                 
                 'bootstrap_servers': 'bootstrap.servers',
                 'client_id': 'client.id',
                 'group_id': 'group.id',
                 'acks': 'request.required.acks',
                 'compression_type': 'compression.type',
                 'retries': 'message.send.max.retries',
                 'linger_ms': 'queue.buffering.max.ms',
                 'request_timeout_ms': 'request.timeout.ms',
                 'retry_backoff_ms': 'retry.backoff.ms',
                 'reconnect_backoff_ms': 'reconnect.backoff.ms',
                 'reconnect_backoff_max_ms': 'reconnect.backoff.max.ms',
                 'max_in_flight_requests_per_connection': 'max.in.flight.requests.per.connection',
                 'metadata_max_age_ms': 'metadata.max.age.ms',
                 'receive_buffer_bytes': 'socket.receive.buffer.bytes',
                 'send_buffer_bytes': 'socket.send.buffer.bytes',
                 'security_protocol': 'security.protocol',
                 'sasl_mechanism': 'sasl.mechanism',
                 'sasl_plain_username': 'sasl.username',
                 'sasl_plain_password': 'sasl.password',
                 'sasl_kerberos_service_name': 'sasl.kerberos.service.name',
               }
  CONFIG_MAP_NULL = [ 'ssl_context', 
                      'socket_options',
                      'batch_size',
                      'partitioner',
                      'buffer_memory',
                      'connections_max_idle_ms',
                      'max_block_ms',
                      'ssl_check_hostname',
                      'api_version',
                      'api_version_auto_timeout_ms',
                      'metrics_reporters',
                      'metrics_num_samples',
                      'metrics_sample_window_ms',
                      'selector',
                      'sasl_kerberos_domain_name',
                      'sasl_oauth_token_provider',
                    ]
  def __init__(self, **configs):
    self._log = logging.getLogger(__name__)
    self.raw_config = configs
    self.cf_config = {}
    self.config = {}
    self.ks = lambda topic, _bytes: bytes(_bytes)
    self.vs = lambda topic, _bytes: bytes(_bytes)
    for k in self.raw_config.keys():
      if (k == 'key_serializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'serialize') or not inspect.isroutine(self.raw_config[k].serialize)):
          self.ks = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.ks = self.raw_config[k].serialize
        self.config[k] = self.ks
      elif (k == 'value_serializer'):
        if not (self.raw_config[k] is None) and (not hasattr(self.raw_config[k], 'serialize') or not inspect.isroutine(self.raw_config[k].serialize)):
          self.vs = lambda topic, _bytes: self.raw_config[k](_bytes)
        else:
          self.vs = self.raw_config[k].serialize
        self.config[k] = self.vs
      elif (k in self.CONFIG_MAP.keys()):
        if not (self.raw_config[k] is None):
          self.cf_config[self.CONFIG_MAP[k]] = self.raw_config[k]
      elif (k in self.CONFIG_MAP_NULL):
        self._log.warning("Warning: Unsupported kafka-python parameter passed to confluent wrapper: %s, %s", k, self.raw_config[k])
      else:
        self.cf_config[k] = self.raw_config[k]
    for oldk,newk in self.CONFIG_MAP.items():
      if newk in self.cf_config.keys():
        self.config[oldk] = self.cf_config[newk]
    super().__init__(self.cf_config)

  def send(self, topic, value=None, key=None, headers=None, partition=0, timestamp_ms=None):
    # wish this were simpler, but the underlying library doesn't like None when no value should be passed.
    if not (headers is None):
      return self.produce(topic,self.vs(topic, value),self.ks(topic, key),partition,lambda a,b: True, timestamp_ms, headers)
    elif not (timestamp_ms is None):
      return self.produce(topic,self.vs(topic, value),self.ks(topic, key),partition,lambda a,b: True, timestamp_ms)
    elif not (partition == 0):
      return self.produce(topic,self.vs(topic, value),self.ks(topic, key),partition)
    elif not (key is None):
      return self.produce(topic,self.vs(topic, value),self.ks(topic, key))
    elif not (value is None):
      return self.produce(topic,self.vs(topic, value))
    else:
      return self.produce(topic)



from threading import Thread
import inspect
import traceback
import pysodium
from time import time,sleep
import msgpack
import logging
from kafkacrypto import TopicPartition,TopicPartitionOffset
import kafka.serializer
from kafkacrypto.base import KafkaCryptoBase
from kafkacrypto.exceptions import KafkaCryptoError, KafkaCryptoSerializeError
from kafkacrypto.message import KafkaCryptoMessage
from kafkacrypto.ratchet import Ratchet
from kafkacrypto.keygenerator import KeyGenerator

class KafkaCrypto(KafkaCryptoBase):
  """Class handling the sending and receiving of encrypted messages.

  Keyword Arguments:
              nodeID (str): Node ID
        kp (KafkaProducer): Pre-initialized KafkaProducer, ready for
                            handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
        kc (KafkaConsumer): Pre-initialized KafkaConsumer, ready ofor      
       	       	       	    handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
         config (str,dict): Either a filename in which configuration
                            data is a stored, or a dict of config
                            parameters. Set to None to load from the
                            default location based on nodeID.
       cryptokey (str,obj): Either a filename in which the crypto
                            private key is stored, or an object
                            implementing the necessary functions 
                            (encrypt_keys, decrypt_keys, sign_spk,
                            load/store_crypto_opaque).
                            Set to None to load from the default
                            location based on nodeID.
            seed (str,obj): Either a filename in which the source
                            ratchet seed is stored, or an object
                            implementing the necessary functions
                            (increment, generate). Set to None to
                            load from the default location based
                            on nodeID.
  """

  #
  # Managed by our process, and protected by _lock since they are also
  # utilized by SerDes subclasses:
  # _tps_updated: set to true anytime a non-management process updates
  #               the list of topic-partitions.
  #         _tps: dict of topic-partitions currently listening to. Needed because
  #               KafkaConsumer does not support incremental assignments, so every
  #               change requires (re)subscribing to all TopicPartitions.
  #               indexed by full topic.
  # _tps_offsets: dict of offsets for each tps. Used since kp,kc may have been
  #               configured without groupIDs, so we have to keep track of our
  #               own offsets.
  #               indexed by full topic.
  # _subs_needed: array of root topics for which new subscriptions are needed.
  #   _subs_last: array of times subscriptions were requested (by root topic).
  #   _cur_pgens: dict of keyidx's for the current producer generator, indexed
  #               by root topic.
  #       _pgens: dict of production encryption key generators, indexed by root
  #               topic, then by 'keyidx', and 'key'/'value'/'secret'.
  #   _new_pgens: Set to true when new pgens were added
  #       _cgens: dict of consumption encryption key generators, indexed by root
  #               topic. Each topic has a list of available generators, indexed
  #               by 'keyidx', and finally 'key'/'value'/'secret'.
  #

  def __init__(self, nodeID, kp, kc, config=None, cryptokey=None, seed=None):
    super().__init__(nodeID, kp, kc, config, cryptokey)
    if (seed is None):
      seed = nodeID + ".seed"
    if (isinstance(seed,(str,))):
      seed = Ratchet(file=seed)
    if (not hasattr(seed, 'increment') or not inspect.isroutine(seed.increment) or not hasattr(seed, 'get_key_value_generators') or not inspect.isroutine(seed.get_key_value_generators)):
      raise KafkaCryptoError("Invalid seed source supplied!")

    self._seed = seed
    self._tps_updated = False
    self._tps = {}
    self._tps_offsets = {}
    self._cur_pgens = {}
    self._pgens = {}
    self._new_pgens = False
    self._cgens = {}
    self._subs_needed = []
    self._subs_last = {}
    self._last_time = time()

    kvs = self._cryptokey.load_crypto_opaque()
    if not (kvs is None):
      if 'pgens' in kvs.keys() or b'pgens' in kvs.keys():
        self._logger.info("Found pgens to load.")
        t = 'pgens'
        if not (t in kvs.keys()):
          t = b'pgens'
        for [root,ki,ks,b] in kvs[t]:
          self._logger.info("Attempting load of root=%s, ki=%s, b=%s at %s", root, ki, b, time())
          if time()-b < self.CRYPTO_MAX_PGEN_AGE:
            if not (root in self._pgens.keys()):
              self._pgens[root] = {}
            self._logger.info("Loaded old key keyindex=%s, birth=%s for root=%s", ki, b, root)
            self._pgens[root][ki] = {}
            self._pgens[root][ki]['birth'] = b
            self._pgens[root][ki]['secret'] = ks
            # stored pgens do not have generators as they should never be used for active production
            # (but secret stays around so lost consumers can catch up)

    self._mgmt_thread = Thread(target=self._process_mgmt_messages,daemon=True)
    self._mgmt_thread.start()

  # Main background processing loop. Must assume that it can "die" at any
  # time, even mid-stride, so ordering of operations to ensure atomicity and
  # durability is critical.
  def _process_mgmt_messages(self):
    while True:
      # First, process messages
      # we are the only thread ever using _kc, _kp, so we do not need the lock to use them
      msgs = self._kc.poll(timeout_ms=self.MGMT_POLL_INTERVAL, max_records=self.MGMT_POLL_RECORDS)
      # but to actually process messages, we need the lock
      for tp,msgset in msgs.items():
        self._lock.acquire()
        for msg in msgset:
          topic = msg.topic
          if (isinstance(topic,(str,))):
            topic = topic.encode('utf-8')
          self._tps_offsets[topic] = msg.offset
          self._logger.debug("Processing message: %s", msg)
          if topic[-len(self.TOPIC_SUFFIX_REQS):] == self.TOPIC_SUFFIX_REQS:
            root = topic[:-len(self.TOPIC_SUFFIX_REQS)]
            # A new receiver: send all available keys
            if root in self._pgens.keys():
              ki = []
              s = []
              for ski,sk in self._pgens[root].items():
                ki.append(ski)
                s.append(sk['secret'])
              if len(ki) > 0:
                k,v = self._cryptokey.encrypt_keys(ki, s, root, msgkey=msg.key, msgval=msg.value)
                if (not (k is None)) or (not (v is None)):
                  self._logger.info("Sending current encryption keys for root=%s to new receiver, msgkey=%s.", root, msg.key)
                  self._kp.send((root + self.TOPIC_SUFFIX_KEYS).decode('utf-8'), key=k, value=v)
                else:
                  self._logger.info("Failed sending current encryption keys for root=%s to new receiver.", root)
              else:
                self._logger.info("No keys for root=%s to send to new receiver.", root)
          elif topic[-len(self.TOPIC_SUFFIX_KEYS):] == self.TOPIC_SUFFIX_KEYS:
            root = topic[:-len(self.TOPIC_SUFFIX_KEYS)]
            # New key(s)
            nks = self._cryptokey.decrypt_keys(root,msgkey=msg.key,msgval=msg.value)
            if not (nks is None):
              for nki,nk in nks.items():
                self._logger.info("Received new encryption key for root=%s, key index=%s, msgkey=%s", root, nki, msg.key)
                # do not clopper other keys that may exist
                if not (root in self._cgens.keys()):
                  self._cgens[root] = {}
                # but do clobber the same topic,keyID entry
                self._cgens[root][nki] = {}
                self._cgens[root][nki]['key'], self._cgens[root][nki]['value'] = KeyGenerator.get_key_value_generators(nk)
                self._cgens[root][nki]['secret'] = nk
          else:
            # unknown object
            self._logger.warning("Unknown topic type in message: %s", msg)
        self._lock.release()
  
      # Second, deal with subscription changes
      self._lock.acquire()
      self._logger.debug("Processing subscription changes.")
      if self._tps_updated == True:
        self._logger.info("Subscriptions changed, adjusting.")
        tpo = []
        for tk in self._tps.keys():
          tpo.append(TopicPartitionOffset(self._tps[tk].topic, self._tps[tk].partition, self._tps_offsets[tk]))
        self._kc.assign_and_seek(tpo)
        self._logger.info("Subscriptions adjusted.")
        if (self._kc.config['group_id'] is None):
          self._logger.info("No group_id, seeking to beginning.")
          self._kc.seek_to_beginning()
        for tk in self._tps.keys():
          if (tk[-len(self.TOPIC_SUFFIX_KEYS):] == self.TOPIC_SUFFIX_KEYS):
            root = tk[:-len(self.TOPIC_SUFFIX_KEYS)]
            if not (root in self._subs_needed):
              self._subs_needed.append(root)
        self._tps_updated = False
      self._lock.release()

      # Third, deal with topics needing subscriptions
      self._lock.acquire()
      self._logger.debug("Process subscriptions.")
      for root in self._subs_needed:
        self._logger.info("(Re)subscribing to root=%s", root)
        if not (root in self._subs_last.keys()) or self._subs_last[root]+self.CRYPTO_SUB_INTERVAL<time():
          k,v = self._cryptokey.signed_epk(root)
          if not (k is None) or not (v is None):
            self._subs_last[root] = time()
            self._logger.info("Sending new subscribe request for root=%s, msgkey=%s", root, k)
            self._kp.send((root + self.TOPIC_SUFFIX_SUBS).decode('utf-8'), key=k, value=v)
          else:
            self._logger.info("Failed to send new subscribe request for root=%s", root)
      self._subs_needed = []
      self._lock.release()

      # Fourth, periodically increment ratchet and prune old keys
      self._logger.debug("Checking ratchet time.")
      self._lock.acquire()
      if self._last_time+self.CRYPTO_RATCHET_INTERVAL < time():
        self._logger.info("Periodic ratchet increment.")
        self._last_time = time()
        # prune
        for root,pgens in self._pgens.items():
          rki = []
          for ki,kv in pgens.items():
            if (kv['birth']+self.CRYPTO_MAX_PGEN_AGE<time()):
              rki.append(ki)
          for ki in rki:
            self._pgens[root].pop(ki)
        # increment
        self._cur_pgens = {}
        self._seed.increment()
      self._lock.release()

      # Fifth, write new producer keys if requested
      self._logger.debug("Checking producer keys.")
      self._lock.acquire()
      if self._new_pgens:
        self._logger.info("(Re)writing producer keys.")
        self._new_pgens = False
        kvs = {}
        kvs['pgens'] = []
        for root,pgens in self._pgens.items():
          for ki,kv in pgens.items():
            kvs['pgens'].append([root,ki,kv['secret'],kv['birth']])
            # stored pgens do not have generators as they should never be used for active production
            # (but secret stays around so lost consumers can catch up)
        self._logger.info("Saving %s old production keys.", len(kvs['pgens']))
        self._cryptokey.store_crypto_opaque(kvs)
      self._lock.release()
  
      # Finally, loop back to poll again
  # end of __process_mgmt_messages

  #
  # We have two internal subclasses that implement the Kafka SerDes interface,
  # and handle new keys/subscriptions as needed.
  #
  class Serializer(kafka.serializer.Serializer):
    def __init__(self, parent, kv):
      self._kv = kv
      self._parent = parent

    def serialize(self, topic, value):
      if (isinstance(topic,(str,))):
        topic = topic.encode('utf-8')
      if value is None:
        return None
      if (isinstance(value,(KafkaCryptoMessage,))):
        value = value.getMessage()
      if (not isinstance(value,(bytes,bytearray))):
        raise KafkaCryptoSerializeError("Passed value is not bytes or a KafkaCryptoMessage")
      root = self._parent.get_root(topic)
      self._parent._lock.acquire()
      try:
        #
        # Slow path: if we don't already have a generator for this topic,
        # we have to make one.
        #
        if not (root in self._parent._cur_pgens.keys()):
          self._parent.get_producer(root)
        # Use generator (new or existing)
        keyidx = self._parent._cur_pgens[root]
        pgen = self._parent._pgens[root][keyidx]
        gen = pgen[self._kv]
        salt = gen.salt()
        key,nonce = gen.generate()
        msg = b'\x01' + msgpack.packb([keyidx,salt,pysodium.crypto_secretbox(value,nonce,key)])
      except Exception as e:
        self._parent._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      finally:
        self._parent._lock.release()
      return msg
 
  class Deserializer(kafka.serializer.Deserializer):
    def	__init__(self, parent, kv, max_key_wait_intervals, key_wait_interval):
      self._kv = kv
      self._parent = parent
      self.MAX_WAIT_INTERVALS = max_key_wait_intervals
      self.WAIT_INTERVAL = key_wait_interval

    def deserialize(self, topic, bytes_):
      if (isinstance(topic,(str,))):
        topic = topic.encode('utf-8')
      if bytes_ is None:
        return None
      root = self._parent.get_root(topic)
      if bytes_[0] != 1:
        return KafkaCryptoMessage.fromBytes(bytes_)
      msg = msgpack.unpackb(bytes_[1:])
      if (len(msg) != 3):
        raise KafkaCryptoSerializeError("Malformed Message!")
      ki = msg[0]
      salt = msg[1]
      msg = msg[2]
      self._parent._lock.acquire()
      i = self.MAX_WAIT_INTERVALS+1
      try:
        while ((not (root in self._parent._cgens.keys()) or not (ki in self._parent._cgens[root].keys())) and i>0):
          i = i-1
       	  ntp =	(root+self._parent.TOPIC_SUFFIX_KEYS)
       	  if not (ntp in self._parent._tps):
       	    self._parent._tps[ntp] = TopicPartition(ntp.decode('utf-8'),0)
            self._parent._tps_offsets[ntp] = 0
            self._parent._tps_updated = True
          else:
            if not (root in self._parent._subs_needed):
              self._parent._subs_needed.append(root)
          if (i > 0):
            self._parent._lock.release()
            sleep(self.WAIT_INTERVAL)
            self._parent._lock.acquire()
        if not (root in self._parent._cgens.keys()) or not (ki in self._parent._cgens[root].keys()):
          self._parent._logger.info("No decryption key found for root=%s, key index=%s. Returning encrypted message.", root, ki)
          raise ValueError
        gen = self._parent._cgens[root][ki][self._kv]
        key,nonce = gen.generate(salt=salt)
        msg = KafkaCryptoMessage(pysodium.crypto_secretbox_open(msg,nonce,key),ipt=True)
      except:
        return KafkaCryptoMessage.fromBytes(bytes_)
      finally:
        self._parent._lock.release()
      return msg

  def getKeySerializer(self):
    return self.Serializer(self,'key')
  def getValueSerializer(self):
    return self.Serializer(self,'value')
  def getKeyDeserializer(self, max_key_wait_intervals=0, key_wait_interval=1):
    return self.Deserializer(self,'key',max_key_wait_intervals,key_wait_interval)
  def getValueDeserializer(self, max_key_wait_intervals=0, key_wait_interval=1):
    return self.Deserializer(self,'value',max_key_wait_intervals,key_wait_interval)

  def get_producer(self,root):
    if (self.MGMT_LONG_KEYINDEX == True):
      ki,ks,kg,vg = self._seed.get_key_value_generators(root, node=self._nodeID)
    else:
      ki,ks,kg,vg = self._seed.get_key_value_generators(root)
    self._cur_pgens[root] = ki
    if not (root in self._pgens):
      self._pgens[root] = {}
    if ki in self._pgens[root] and (not ('key' in self._pgens[root][ki].keys()) or not ('value' in self._pgens[root][ki].keys())):
      self._logger.critical("Potential key reuse scenario detected! Incrementing ratchet for safety.")
      self._seed.increment()
    elif not (ki in self._pgens[root].keys()):
      self._new_pgens = True
      self._pgens[root][ki] = {}
      self._pgens[root][ki]['key'] = kg
      self._pgens[root][ki]['value'] = vg
      self._pgens[root][ki]['secret'] = ks
      self._pgens[root][ki]['birth'] = time()
      ntp = (root+self.TOPIC_SUFFIX_REQS)
      if not (ntp in self._tps):
        self._tps[ntp] = TopicPartition(ntp.decode('utf-8'),0)
        self._tps_offsets[ntp] = 0
        self._tps_updated = True
      self._logger.info("Got new producer key for root=%s. New key index=%s", root, ki)

